
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

typedef struct {
		double	x1,y1,z1,x2,y2,z2,x3,y3,z3,
			nx,ny,nz,id12,ox12,oy12,oz12,
				 id23,ox23,oy23,oz23,
				 id31,ox31,oy31,oz31;
	} SI;

#define D(x,y,z)	sqrt(x*x+y*y+z*z)

void	SetUpSI(double x1,double y1,double z1,
		double x2,double y2,double z2,
		double x3,double y3,double z3,SI *si)
{
	double	x12=x2-x1,y12=y2-y1,z12=z2-z1,id12=1.0/D(x12,y12,z12),
		x23=x3-x2,y23=y3-y2,z23=z3-z2,id23=1.0/D(x23,y23,z23),
		x31=x1-x3,y31=y1-y3,z31=z1-z3,id31=1.0/D(x31,y31,z31),
		a=z12*y31-y12*z31,b=x12*z31-z12*x31,c=y12*x31-x12*y31,
		d=D(a,b,c),nx=a/d,ny=b/d,nz=c/d;

	si->x1=x1; si->y1=y1; si->z1=z1;
	si->x2=x2; si->y2=y2; si->z2=z2;
	si->x3=x3; si->y3=y3; si->z3=z3;
	si->nx=nx; si->ny=ny; si->nz=nz;
	si->id12=id12; si->ox12=id12*(y12*nz-z12*ny);
		       si->oy12=id12*(z12*nx-x12*nz);
		       si->oz12=id12*(x12*ny-y12*nx);
	si->id23=id23; si->ox23=id23*(y23*nz-z23*ny);
		       si->oy23=id23*(z23*nx-x23*nz);
		       si->oz23=id23*(x23*ny-y23*nx);
	si->id31=id31; si->ox31=id31*(y31*nz-z31*ny);
		       si->oy31=id31*(z31*nx-x31*nz);
		       si->oz31=id31*(x31*ny-y31*nx);
}

#define EPS	2.22045e-16

static double	LI(double d,double e)
{
	double	r=(d-1.0)/(d+1.0);

	return (r<EPS)?0.0:-log(r)*e;
}

void	SurfaceIntegral(unsigned long F,SI *si,
			double x0,double y0,double z0,
			double *P,double *X,double *Y,double *Z)
{
	double	x1,y1,z1,d1,x2,y2,z2,d2,x3,y3,z3,d3,f,g,
		p=0.0,x=0.0,y=0.0,z=0.0;

	for (; F!=0; F--, si++) {
	    x1=si->x1-x0; y1=si->y1-y0; z1=si->z1-z0; d1=D(x1,y1,z1);
	    x2=si->x2-x0; y2=si->y2-y0; z2=si->z2-z0; d2=D(x2,y2,z2);
	    x3=si->x3-x0; y3=si->y3-y0; z3=si->z3-z0; d3=D(x3,y3,z3);
	    f=x1*si->nx+y1*si->ny+z1*si->nz;
	    g=LI(si->id12*(d1+d2),x1*si->ox12+y1*si->oy12+z1*si->oz12)
	     +LI(si->id23*(d2+d3),x2*si->ox23+y2*si->oy23+z2*si->oz23)
	     +LI(si->id31*(d3+d1),x3*si->ox31+y3*si->oy31+z3*si->oz31)
	     -f*2.0*atan2(x1*y2*z3+x2*y3*z1+x3*y1*z2
			 -x3*y2*z1-x1*y3*z2-x2*y1*z3,
			  d1*d2*d3+d1*(x2*x3+y2*y3+z2*z3)
				  +d2*(x3*x1+y3*y1+z3*z1)
				  +d3*(x1*x2+y1*y2+z1*z2));
	    p+=(f*g); x-=(si->nx*g); y-=(si->ny*g); z-=(si->nz*g);
	}
	*P=p/2.0; *X=x; *Y=y; *Z=z;
}

#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_TYPE	HANDLE
#define FUNCTION_TYPE	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(THREAD,FUNCTION,ARGUMENT)	\
	(THREAD=(HANDLE) \
		_beginthreadex(NULL,0,FUNCTION,(void *)ARGUMENT,0,NULL))==0

#define TERM_THREAD(THREAD)	\
	WaitForSingleObject(THREAD,INFINITE)==WAIT_FAILED || \
	CloseHandle(THREAD)==0
#else
#include <pthread.h>

#define THREAD_TYPE	pthread_t
#define FUNCTION_TYPE	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(THREAD,FUNCTION,ARGUMENT)	\
	pthread_create(&(THREAD),NULL,FUNCTION,(void *)ARGUMENT)

#define TERM_THREAD(THREAD)	pthread_join(THREAD,NULL)
#endif

static int		M=1;
static unsigned long	F,N;
static double		*X,*Y,*Z,*P;
static SI		*si;

static FUNCTION_TYPE	Thread(void *arg)
{
	size_t	i;

	for (i=(size_t)arg; i<N; i+=M)
	    SurfaceIntegral(F,si,X[i],Y[i],Z[i],P+i,X+i,Y+i,Z+i);

	return RETURN_VALUE;
}

#define LEN	2048
#define INC	2048
#define SOD	sizeof(double)

int	main(int argc,char **argv)
{
	char		*e,str[LEN],
			*fnf=" : file not found.",
			*dre="data read error.",
			*mae="memory allocation error.",
			*mrae="memory re-allocation error.";
	FILE		*file;
	unsigned long	V,i,v1,v2,v3;
	double		v,x,y,z,x1,y1,z1,x2,y2,z2,x3,y3,z3,w,p1,p2;
	THREAD_TYPE	*T;

	if (argc!=2 && argc!=3) Error("usage : vf_m vf.txt {point.txt}");

	if ((e=getenv("THREADS"))!=NULL &&
	    (M=atoi(e))<=0) Error("bad number of THREADS.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"r"))==NULL) {
		(void)fputs(argv[1],stderr); Error(fnf);
	    }

	if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
	    sscanf(str,"%lu %lu",&V,&F)!=2 ||
	    V==0 || F==0) Error("bad data number.");

	if ((X=(double *)malloc(SOD*V))==NULL ||
	    (Y=(double *)malloc(SOD*V))==NULL ||
	    (Z=(double *)malloc(SOD*V))==NULL ||
	    (si=(SI *)malloc(sizeof(SI)*F))==NULL) Error(mae);

	for (i=0; i<V; i++)
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lf %lf %lf",X+i,Y+i,Z+i)!=3) Error(dre);

	v=x=y=z=0.0;
	for (i=0; i<F; i++) {
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lu %lu %lu",&v1,&v2,&v3)!=3 ||
		v1>=V || v2>=V || v3>=V) Error(dre);

	    SetUpSI(x1=X[v1],y1=Y[v1],z1=Z[v1],
		    x2=X[v2],y2=Y[v2],z2=Z[v2],
		    x3=X[v3],y3=Y[v3],z3=Z[v3],si+i);

	    v+=(w=x1*y2*z3+x2*y3*z1+x3*y1*z2-x3*y2*z1-x1*y3*z2-x2*y1*z3);
	    x+=(w*(x1+x2+x3)); y+=(w*(y1+y2+y3)); z+=(w*(z1+z2+z3));
	}
	if (file!=stdin) (void)fclose(file);

	w=v*4.0;
	(void)printf("%lf\t%lf\t%lf\t%lf\t",v/6.0,x/=w,y/=w,z/=w);

	SurfaceIntegral(F,si,x,y,z,&w,&x,&y,&z);
	(void)printf("%lf\t%lf\t%lf\t%lf\n",w,x,y,z);

	if (argc==2)
	    N=V;
	else {
	    if (argv[2][0]=='-')
		if (argv[2][1]=='-') {
		    if (argv[2][2]=='-')
			for (i=0; i<V; i++) (void)putchar('\n');

		    if ((X=(double *)realloc(X,SOD*F))==NULL ||
			(Y=(double *)realloc(Y,SOD*F))==NULL ||
			(Z=(double *)realloc(Z,SOD*F))==NULL) Error(mrae);

		    for (N=0; N<F; N++) {
			X[N]=(si[N].x1+si[N].x2+si[N].x3)/3.0;
			Y[N]=(si[N].y1+si[N].y2+si[N].y3)/3.0;
			Z[N]=(si[N].z1+si[N].z2+si[N].z3)/3.0;
		    }
		    goto Break;
		}
		else
		    file=stdin;
	    else
		if ((file=fopen(argv[2],"r"))==NULL) {
		    (void)fputs(argv[2],stderr); Error(fnf);
		}

	    N=0;
	    while (fgets(str,LEN,file)!=NULL && str[strlen(str)-1]=='\n' &&
		   sscanf(str,"%lf %lf %lf",&x,&y,&z)==3) {
		if (V==N) {
		    V+=INC;
		    if ((X=(double *)realloc(X,SOD*V))==NULL ||
			(Y=(double *)realloc(Y,SOD*V))==NULL ||
			(Z=(double *)realloc(Z,SOD*V))==NULL) Error(mrae);
		}
		X[N]=x; Y[N]=y; Z[N]=z; ++N;
	    }
	    if (N==0) Error("no point data.");

	    if ((X=(double *)realloc(X,SOD*N))==NULL ||
		(Y=(double *)realloc(Y,SOD*N))==NULL ||
		(Z=(double *)realloc(Z,SOD*N))==NULL) Error(mrae);

	    if (file!=stdin) (void)fclose(file);
	}
Break:;
	if ((P=(double *)malloc(SOD*N))==NULL ||
	    (T=(THREAD_TYPE *)malloc(sizeof(THREAD_TYPE)*M))==NULL) Error(mae);

	for (i=1; i<M; i++)
	    if (INIT_THREAD(T[i],Thread,(size_t)i))
		Error("multi-threading initialization error.");

	(void)Thread(0);

	for (i=1; i<M; i++)
	    if (TERM_THREAD(T[i])) Error("multi-threading termination error.");

	for (i=0; i<N; i++) {
	    (void)printf("%lf\t%lf\t%lf\t%lf\n",P[i],X[i],Y[i],Z[i]);

	    if (i==0) {
		p1=p2=P[0]; x1=x2=X[0]; y1=y2=Y[0]; z1=z2=Z[0];
	    }
	    else {
		if (p1>P[i]) p1=P[i]; else if (p2<P[i]) p2=P[i];
		if (x1>X[i]) x1=X[i]; else if (x2<X[i]) x2=X[i];
		if (y1>Y[i]) y1=Y[i]; else if (y2<Y[i]) y2=Y[i];
		if (z1>Z[i]) z1=Z[i]; else if (z2<Z[i]) z2=Z[i];
	    }
	}
	(void)fprintf(stderr,"%lf\t%lf\n"
			     "%lf\t%lf\n"
			     "%lf\t%lf\n"
			     "%lf\t%lf\n",p1,p2,x1,x2,y1,y2,z1,z2);

	return 0;
}
